%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Brigitta Baranyai <baranyai.brigitta26@gmail.com>

-module(reflib_secure).
-vsn("$Rev: 16814 $").

-export([get_unsecure_calls/1, get_calls_for_os/1, get_calls_for_port/1, 
	 get_calls_for_file/1, get_calls_for_unstable_funs/1, get_calls_for_nif/1, 
	 get_calls_for_port_driver/1, get_calls_for_old_crypto_api/1,
	 get_calls_for_old_crypto_api_help/1, unsecure_compile_and_load_operations/1, 
	 unsecure_process_linkage/1, unsecure_prioritization/1, get_unsecure_concurrent_calls/1,
	 unsecure_ets_traversal/1, unsafe_network/1, get_unsecure_interoperability/1,
	 get_unsecure_xml_usage/1, check_unsecure_communication/1]).

-include("lib.hrl").

-define(os_call, [{os, [cmd, putenv]}]).
-define(port_creation_call, [{erlang, [open_port]}]).
-define(nif_call, [{erlang, [load_nif]}]).
-define(port_driver_call, [{erl_ddll, [load, load_driver, reload, reload_driver, try_load]}]).
-define(file_operation, [{file, [consult, path_consult, eval, path_eval, path_script, script]}]).
-define(unstable_atom_call, [{erlang, [binary_to_atom, binary_to_term, list_to_atom]},
							 {http_uri, [parse]}]).
-define(crypto_call, [{crypto, [sign, verify]}]).
-define(old_crypto_api_call, [{crypto, [block_encrypt, block_decrypt, stream_init, stream_encrypt, 
							stream_decrypt, cmac, hmac, hmac_init, hmac_update, hmac_final, 
							hmac_final_n, poly1305]}]).
-define(compile_and_load_call, [{compile, [file, noenv_file]}, 
								{code, [atomic_load, load_abs, load_binary, load_file]}]).
-define(process_linkage_call, [{erlang, [link]}]).
-define(process_flag_call, [{erlang, [process_flag]}]).
-define(ets_traversal_call, [{ets, [first, next, prev, last, match, match_delete, match_object, select, 
									select_count, select_delete, select_replace, select_reverse,
									tab2file, tab2list]}]).
-define(network_call, [{net_kernel, [allow, connect_node, start]},
						{net_adm, [host_file, world]}]).
-define(xml_usage, [{xmerl_scan, [file, string]}, 
					{xmerl_sax_parser, [file, stream]}]).
-define(unsecure_communication, [{ssl, [connect, handshake, handshake_continue, listen, ssl_accept]}]).


get_unsecure_calls(Fun) ->
	Result = [ get_calls_for_commons(Fun), get_calls_for_port(Fun), get_unsecure_concurrent_calls(Fun),
			get_calls_for_unstable_funs(Fun), unsecure_compile_and_load_operations(Fun), 
			get_calls_for_old_crypto_api(Fun), get_unsecure_xml_usage(Fun), 
			check_unsecure_communication(Fun) ],
	lists:flatten(Result).

get_unsecure_interoperability(Fun) ->
	lists:flatten([ get_calls_for_port(Fun), get_calls_for_nif(Fun), get_calls_for_port_driver(Fun) ]).

get_unsecure_concurrent_calls(Fun) ->
	lists:flatten([ unsecure_process_linkage(Fun), unsecure_prioritization(Fun), unsecure_ets_traversal(Fun) ]).

get_calls_for_commons(Fun) ->
	Types = ?os_call ++ ?nif_call ++ ?port_driver_call ++ ?file_operation ++ ?network_call,
	DataFlowTuples = analyze_funs(Fun, Types, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge(get_funs_without_body(NonSafeDataFlowTuples), get_exported_funs(NonSafeDataFlowTuples)).

get_calls_for_os(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?os_call, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge(get_funs_without_body(NonSafeDataFlowTuples), get_exported_funs(NonSafeDataFlowTuples)).

get_calls_for_port(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?port_creation_call, fun get_children_for_port/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge(get_funs_without_body(NonSafeDataFlowTuples), get_exported_funs(NonSafeDataFlowTuples)).

get_calls_for_nif(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?nif_call, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge(get_funs_without_body(NonSafeDataFlowTuples), get_exported_funs(NonSafeDataFlowTuples)).

get_calls_for_port_driver(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?port_driver_call, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge(get_funs_without_body(NonSafeDataFlowTuples), get_exported_funs(NonSafeDataFlowTuples)).

get_calls_for_file(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?file_operation, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge(get_funs_without_body(NonSafeDataFlowTuples), get_exported_funs(NonSafeDataFlowTuples)).	

get_calls_for_unstable_funs(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?unstable_atom_call, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	FunsWithSafeOption = [ {O, {P, D}} || {O, {P, D}} <- NonSafeDataFlowTuples, 
								contains_specific_options(D, fun get_options/1, fun safe_options/1) ],
	FunsWithoutSafeOption = NonSafeDataFlowTuples -- FunsWithSafeOption,
	lists:merge(get_funs_without_body(FunsWithoutSafeOption), get_exported_funs(FunsWithoutSafeOption)).

safe_options(El) when is_atom(El), El =:= safe ->
	true;
safe_options({scheme_validation_fun, _}) ->
	true;
safe_options(El) when is_atom(El), El =/= safe ->
	false;
safe_options(_) ->
	false.

get_calls_for_old_crypto_api(Fun) ->
	get_calls_for(Fun, lists:merge(?old_crypto_api_call, ?crypto_call)).

get_calls_for_old_crypto_api_help(Fun) ->
	Expressions = get_calls_for(Fun, lists:merge(?old_crypto_api_call, ?crypto_call)),
	ExprFunTuple = [ {E, ?Fun:name(hd(?Query:exec(E, ?Expr:functions())))} || E <- Expressions ],
	[ {E, list_to_atom(string:concat("Please use the following functions instead: ", 
							string:join(handle_old_crypto_api(F), ", ")))} 
				|| {E, F} <- ExprFunTuple ].

handle_old_crypto_api(Name) ->
	case Name of
		block_encrypt -> ["crypto_one_time", "crypto_one_time_aead",
					      "crypto_init", "crypto_update",
					      "crypto_dyn_iv_init", "crypto_dyn_iv_update"];
		block_decrypt -> ["crypto_one_time", "crypto_one_time_aead",
					      "crypto_init", "crypto_update",
					      "crypto_dyn_iv_init", "crypto_dyn_iv_update"];
		stream_init -> ["crypto_one_time", "crypto_one_time_aead",
					      "crypto_init", "crypto_update",
					      "crypto_dyn_iv_init", "crypto_dyn_iv_update"];
		stream_encrypt -> ["crypto_one_time", "crypto_one_time_aead",
					      "crypto_init", "crypto_update", 
					      "crypto_dyn_iv_init", "crypto_dyn_iv_update"];
		stream_decrypt -> ["crypto_one_time", "crypto_one_time_aead",
					      "crypto_init", "crypto_update",
					      "crypto_dyn_iv_init", "crypto_dyn_iv_update"];
		cmac -> ["mac", "macN"];
		hmac -> ["mac", "macN"];
		hmac_init -> ["mac_init"];
		hmac_update -> ["mac_update"];
		hmac_final -> ["mac_final"];
		hmac_final_n -> ["mac_finalN"];
		poly1305 -> ["mac", "macN"];
		sign -> ["public_key:sign"];
		verify -> ["public_key:verify"];
		_Else -> []
	end.

unsecure_compile_and_load_operations(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?compile_and_load_call, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge3(get_funs_without_body(NonSafeDataFlowTuples), 
				check_for_loaded_modules(NonSafeDataFlowTuples),
		    	get_exported_funs(NonSafeDataFlowTuples)).

unsecure_process_linkage(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?process_linkage_call, fun get_all_children_for/1),
	ConvertedFunParamTuples = [ {O, ?Query:exec(D, ?Expr:functions())} || {O, {_, D}} <- DataFlowTuples ],
	[ O || {O, P} <- ConvertedFunParamTuples, P =/= [] andalso 
			(?Mod:name(hd(?Query:exec(P, ?Fun:module()))) == erlang) andalso 
			(?Fun:name(hd(P)) == spawn) ].

unsecure_prioritization(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?process_flag_call, fun get_all_children_for/1),
	ConvertedFunParamTuples = [ {O, get_term_of_params(D)} || {O, {_, D}} <- DataFlowTuples ],
	[ O || {O, P} <- ConvertedFunParamTuples, P =/= [] andalso lists:member(priority, P) ].

unsecure_ets_traversal(Fun) ->
	MatchingExpressions = get_calls_for(Fun, ?ets_traversal_call),
	FunParamTuples = [ {E, get_expression_params_for(E, ?ets_traversal_call, fun get_all_children_for/1)} 
					|| E <- MatchingExpressions ],
	DataFlowTuples = get_origins(FunParamTuples, [{back, true}]),
	Functions = [ {O, {P, ?Query:exec(D, ?Query:seq([?Expr:clause(), ?Clause:form(), ?Form:func()]))}}
					|| {O, {P, D}} <- DataFlowTuples ],
	FunctionCalls = [ {O, {P, get_embedded_funs(F)}} || {O, {P, F}} <- Functions ],
	SafeFixTableExpressions = [ {O, {P, get_calls_for(F, [{ets, [safe_fixtable]}])}} 
					|| {O, {P, F}} <- FunctionCalls ],
	SafeFixTableExprWithValues = [ {O, {P, get_expression_params_for(F, [{ets, [safe_fixtable]}], 
					fun get_all_children_for/1)}} || {O, {P, F}} <- SafeFixTableExpressions ],
	IsTrue = fun(E) -> E == true end,
	MatchingExprWithSafeFixTableExpr = [ O || {O, {_, F}} <- SafeFixTableExprWithValues, 
					lists:any(IsTrue, lists:flatten(check_for_params(F))) ],
	[ E || E <- MatchingExpressions, not lists:member(E, MatchingExprWithSafeFixTableExpr) ].

get_embedded_funs([]) -> [];
get_embedded_funs(Funs) ->
	FunctionCalls = [ ?Query:exec(F, ?Query:all([?Fun:funcalls(), ?Dynfun:dyncalls()])) 
					|| F <- Funs ],
	Set = sets:from_list(lists:flatten(FunctionCalls)),
	Result = [ F || F <- sets:to_list(Set), not lists:member(F, Funs)],
	lists:umerge(Funs, get_embedded_funs(Result)).

check_for_params(ParamList) ->
	[ get_term_of_params(A2) || [_|A2] <- ParamList ].

unsafe_network(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?network_call, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	lists:merge(get_funs_without_body(NonSafeDataFlowTuples), get_exported_funs(NonSafeDataFlowTuples)).

get_unsecure_xml_usage(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?xml_usage, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	FunsWithSafeOption = [ {O, {P, D}} || {O, {P, D}} <- NonSafeDataFlowTuples, 
								contains_specific_options(D, fun get_options_for_xml/1, 
										fun entity_expansion_check/1) ],
	FunsWithoutSafeOption = NonSafeDataFlowTuples -- FunsWithSafeOption,
	lists:merge(get_funs_without_body(FunsWithoutSafeOption), get_exported_funs(FunsWithoutSafeOption)).

get_options_for_xml(Expr) ->
	case ?Expr:type(Expr) of 
		tuple -> 
			lists:zip(get_term_of_params(?Query:exec(Expr, ?Expr:child(1))), 
					?Query:exec(Expr, ?Expr:child(2)));
		_ ->
			Expr
	end.

entity_expansion_check({event_fun, EventFun}) ->
	DataFlow = ?Dataflow:reach([EventFun], [{back, true}]),
	TupleExprs = [ ?Query:exec(D, ?Expr:clauses()) || D <- DataFlow ],
	OptionTuples = [ ?Query:exec(T, ?Clause:pattern(1)) || T <- lists:flatten(TupleExprs), T =/= [] ],
	TupleValues = [ hd(get_term_of_params(?Query:exec(T, ?Expr:child(1)))) 
					|| T <- lists:flatten(OptionTuples) ],
	check_for_entity_event(hd(TupleValues)).

check_for_entity_event(internalEntityDecl) ->
	true;
check_for_entity_event(externalEntityDecl) ->
	true;
check_for_entity_event(_) ->
	false.

check_unsecure_communication(Fun) ->
	DataFlowTuples = analyze_funs(Fun, ?unsecure_communication, fun get_all_children_for/1),
	NonSafeDataFlowTuples = get_unsafe_funs(DataFlowTuples),
	[ O || {O, {_, D}} <- NonSafeDataFlowTuples, 
					contains_specific_options(D, fun get_options/1, fun unsafe_options/1) ].

contains_specific_options(DataFlow, OptionGetter, OptionChecker) ->
	TupleExprs = [ ?Query:exec(D, ?Expr:children()) || D <- DataFlow ],
	OptionTuples = [ ?Query:exec(T, ?Expr:children()) || T <- lists:flatten(TupleExprs) ],
	TupleValues = [ OptionGetter(T) || T <- lists:flatten(OptionTuples) ],
	UnsafeOptions = [ OptionChecker(T) || T <- lists:flatten(TupleValues) ],
	IsTrue = fun(E) -> E == true end,
	lists:any(IsTrue, UnsafeOptions).

get_options(Expr) ->
	case ?Expr:type(Expr) of 
		tuple -> 
			lists:zip(get_term_of_params(?Query:exec(Expr, ?Expr:child(1))), 
					get_term_of_params(?Query:exec(Expr, ?Expr:child(2))));
		atom ->
			?Expr:value(Expr);
		_ ->
			Expr
	end.

unsafe_options({padding_check, false}) ->
	true;
unsafe_options({beast_mitigation, disabled}) ->
	true;
unsafe_options({fallback, true}) ->
	true;
unsafe_options({dh, _}) ->
	true;
unsafe_options({dhfile, _}) ->
	true;
unsafe_options({client_renegotiation, true}) ->
	true;
unsafe_options({verify, verify_none}) ->
	true;
unsafe_options({crl_check, false}) ->
	true;
unsafe_options({_, _}) ->
	false;
unsafe_options(_) ->
	false.

analyze_funs(Fun, Type, FunDefForParam) ->
	FunExpressions = get_calls_for(Fun, Type),
	FunParamTuples = [ {E, get_expression_params_for(E, Type, FunDefForParam)} || E <- FunExpressions ],
	get_origins(FunParamTuples, [{back, true}]).

get_calls_for(Fun, ModFnListTuples) ->
	Expressions = ?Query:exec(Fun, ?Query:seq([?Fun:definition(), ?Form:clauses(), ?Clause:exprs()])),
	Functions = [ ?Query:exec(E, ?Expr:functions()) || E <- Expressions ],
	Tuples = lists:zip(Expressions, Functions),
	get_functions_for(Tuples, ModFnListTuples).

get_origins(FunParamTuples, Opt) ->
	%% TODO: make it recursive
	FirstDataFlowTuples = [ {O, {P, ?Dataflow:reach(P, Opt)}} || {O, P} <- FunParamTuples ],
	SelDataFlowTuples = [ {O, {P, {D, ?Query:exec(D, [{sel, back}])}}} 
				|| {O, {P, D}} <- FirstDataFlowTuples ],
	SecondDataFlowTuples = [ {O, {P, lists:umerge(D, ?Dataflow:reach(S, Opt))}} 
				|| {O, {P, {D, S}}} <- SelDataFlowTuples ],
	SecondSelDataFlowTuples = [ {O, {P, {D, ?Query:exec(D, [{sel, back}])}}} 
				|| {O, {P, D}} <- SecondDataFlowTuples ],
	[ {O, {P, lists:umerge(D, ?Dataflow:reach(S, Opt))}} 
				|| {O, {P, {D, S}}} <- SecondSelDataFlowTuples ].

get_expression_params_for([], _, _) -> [];
get_expression_params_for(ExprList, Type, FunDef) when is_list(ExprList) ->
	[ get_expression_params_for(Expr, Type, FunDef) || Expr <- ExprList ];
get_expression_params_for(Expr, Type, FunDef) ->
	case ?Expr:type(Expr) of 
		application ->
			SubExpr = ?Query:exec(Expr, ?Expr:deep_sub()),
			FilteredSubExpr = [ Sub || Sub <- SubExpr, ?Expr:type(Sub) =:= application andalso (is_same_fun(Sub, Type)) ],
			FunDef(FilteredSubExpr);
		match_expr ->
			get_expression_params_for(hd(?Query:exec(Expr, ?Expr:child(2))), Type, FunDef);
		cons ->
			?Query:exec(Expr, ?Query:all(?Query:seq(?Expr:child(1), ?Expr:children()), ?Expr:child(2)));
		_ ->
			?Query:exec(Expr, ?Expr:children())
	end.

is_same_fun(Expr, ModFnListTuples) ->
	StringExpr = hd(string:split(?Syn:flat_text2(Expr), "(")),
	IsSameFuns = [ [ string:find(StringExpr, atom_to_list(F) , leading) =/= nomatch 
					|| F <- FnList ] || {_, FnList}  <- ModFnListTuples ],
	IsTrue = fun(E) -> E == true end,
	Result = lists:any(IsTrue, lists:flatten(IsSameFuns)),
	Result.

get_children_for_port(Expr) ->
	?Query:exec(Expr, ?Query:seq([?Expr:child(2),
							?Expr:child(1),
							?Expr:child(2)])).

get_all_children_for(Expr) ->
	?Query:exec(Expr, ?Query:seq([?Expr:child(2), ?Expr:children()])).

check_for_loaded_modules(Tuples) ->
	LoadedModules =  [ M || {M, _} <- code:all_loaded() ],
	IsTrue = fun(E) -> E == true end,
	[ O || {O, {_, D}} <- Tuples, 
				lists:any(IsTrue, is_member(get_term_of_params(D), LoadedModules)) ].

get_funs_without_body([]) -> [];
get_funs_without_body(Tuples) ->
	Functions = [ {O, {P, ?Query:exec(D, ?Expr:functions())}} || {O, {P, D}} <- Tuples ],
	FunctionBodies = [ {O, {P, [ ?Query:exec(F, ?Query:seq([?Fun:definition(),
															?Form:clauses(), 
								  							?Clause:body()])) || F <- Fun ]}}
			   	|| {O, {P, Fun}} <- Functions, Fun =/= [] ],
	FunctionWithoutBody = [ O || {O, {_, B}} <- FunctionBodies, lists:member([], B) ],
	OriginalDataFlows = [ ?Dataflow:reach(P, [{back, true}]) || {_, {P, _}} <- Tuples ],
	IsMember = fun(E) -> lists:member(E, lists:flatten(OriginalDataFlows)) end,
	FunctionWithBodies = [ {O, {P, lists:filter(IsMember, B)}} || {O, {P, B}} <- FunctionBodies, B =/= [] ],
	lists:merge(FunctionWithoutBody, get_funs_without_body(FunctionWithBodies)).

get_exported_funs(Tuples) -> 
	Functions = [ {O, {P, ?Query:exec(D, ?Query:seq([?Expr:clause(), 
							 ?Clause:form(), 
							 ?Form:func()]))}} 
			  || {O, {P, D}} <- Tuples ],
	ExportedFuns = [ {O, {P, F}} || {O, {P, F}} <- Functions, ?Fun:is_exported(F) ],
	FuncWithParam = [ {O, {get_term_of_params(P), 
			       get_term_of_params(?Query:exec(F, ?Query:seq([?Fun:definition(),
									     ?Form:clauses(),
									     ?Clause:patterns()])))}} 
			  || {O, {P, F}} <- ExportedFuns ],
	IsTrue = fun(E) -> E == true end,
	[ O || {O, {P, A}} <- FuncWithParam, lists:any(IsTrue, is_member(P, A)) ].

get_unsafe_funs(Tuples) ->
	Functions = [ {O, ?Query:exec(D, ?Expr:functions())} || {O, {_, D}} <- Tuples ],
	SafeFuns = lists:flatten(?Syn:get_env(safe_funs)),
	IsTrue = fun(E) -> E == true end,
	IsSafe = fun(F) -> lists:any(IsTrue, is_funs_equals([F], SafeFuns)) end,
	NonSafeFuns = [ E || {E, F} <- Functions, not lists:any(IsSafe, F) ],
	[ {E, {P, D}} || {E, {P, D}} <- Tuples, lists:member(E, lists:flatten(NonSafeFuns)) ].

is_member(P1, P2) ->
	[ lists:member(Item, P2) || Item <- P1, Item =/= undefined ].

get_term_of_params(Params) ->
	[ ?Expr:value(P) || P <- Params ].

get_functions_for(Tuples, ModFnListTuples) ->
	IsTrue = fun(E) -> E == true end,
	[ E || {E, F} <- Tuples, lists:any(IsTrue, is_funs_equals(F, ModFnListTuples)) ].

is_funs_equals(Funs, ModFnListTuples) ->
	IsTrue = fun(E) -> E == true end,
	[ Funs =/= [] andalso lists:any(IsTrue, is_fun_equals(Funs, Mod, FnList))
				|| {Mod, FnList} <- ModFnListTuples ].

is_fun_equals(TargetFuns, Mod, FnList) ->
	[ lists:member(?Fun:name(F), FnList) andalso 
		(?Mod:name(hd(?Query:exec(F, ?Fun:module()))) == Mod) || F <- TargetFuns ].
